/*******************************************************\
* reads the data and runs the discoverers analysis      *
* see http://arxiv.org/abs/1509.01477 for details       *
* matus.medo@unifr.ch, June 2016                        *
\*******************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cdf.h>


#define TOPD            20                      // how many best-discovering users will be shown
#define EPSILON         1e-8                    // epsilon for floating-point comparisons
#define REP_BOOT        10000                   // how many bootstrap repetitions are run to determine the average maximal bootstrapped surprisal
#define compact_real_profile    0               // whether to save the real bootstrap profile in a compact form (1 - only ranks where the surprisal value changes are saved) or not (0)


typedef struct                                  // structure for the data
{
  int user;
  int object;
  int time;
} datastruct;


int U, O, L, TMAX;                              // number of users, objects, and links and the maximal time stamp
int degD = 5;                                   // the first degD users who collect a target object are awarded with discoveries
double fracD = 0.01;                            // fraction of objects which are there to be discovered
datastruct *data;                               // user-object data with time
int *udeg, *odeg, *discov;                      // user and object degree values; user number of discoveries
double pD;                                      // discovery probability per link
double *probNull;                               // probability of at least d_i discoveries out of k_i attempts under the null model (user surprisal)
double *surp;                                   // user surprisal: surp = -ln(probNull)
int *tau;                                       // day of the first occurance for all objects
double sum_max_s_real, sum_max_s_boot;          // to compute the average maximal surprisal in real data and bootstrap
const gsl_rng_type *T;                          // gsl constant
gsl_rng *r;                                     // gsl constant


void init(){
  struct stat st;
  if (stat("data_analysis", &st) != 0) mkdir("data_analysis", 0777);                    // if it does not exist already, create the output directory for data analysis

  gsl_rng_env_setup();
  gsl_rng_default_seed = 0;
  T = gsl_rng_default;
  r = gsl_rng_alloc(T);
}


void print_usage(){
  printf("usage: analyze_data dataset [degD] [fracD]\n");
  printf("       optional parameter degD specifies how many first users are awarded a discovery (integer; default: 5)\n");
  printf("       optional parameter fracD specifies the fraction of most popular items for which discoveries are awarded (floating-point; default: 0.01)\n");
}


double get_Null_prob(int D, int deg, double probD)                                      // probability of at least D discoveries out of deg attempts if the probability of each discovery is probD
{
  if (D == 0) return(1);
  else return(gsl_cdf_binomial_Q(D - 1, probD, deg));
}


double do_bootstrap(char dataset[])
{
  int i, j, numD, rep_step;
  double surprisal, max_s, sum_max_s = 0;

  printf("running bootstrap: "); fflush(stdout);
  if (REP_BOOT < 100) rep_step = 1;
  else rep_step = REP_BOOT / 100;
  for (i = 0; i < REP_BOOT; i++){
    if ((i + 1) % rep_step == 0){printf("%d%% ", (i + 1) / rep_step); fflush(stdout);}
    max_s = -1;
    for (j = 0; j < U; j++){                                    // generate bootstrapped surprisal values
      numD = gsl_ran_binomial(r, pD, udeg[j]);                  // number of discoveries is drawn from the binomial distribution
      surprisal = -log(get_Null_prob(numD, udeg[j], pD));       // corresponding surprisal value
      if (surprisal > max_s) max_s = surprisal;
    }
    sum_max_s += max_s;
  }
  printf("\n\n");
  return (sum_max_s / REP_BOOT);
}


void load_data(char dataset[]){                                                         // load the input bipartite user-item data
  int a, i, j, n, user, object, time, counter, counterD, bestD, whomax;
  int first_line = 1, prev_time = 0;
  int *odegnow, *odeginc, timenow, mindegD, numD;
  int maxUdeg = 0, maxOdeg = 0;
  double max_s, prev_val;
  int *sum_k_collect, *intop, *used;                                                    // the total degree of collected items and whether a user is already among the top ones (1) or not (0)
  char flag, one_line[2000], oname[100];
  FILE *input, *output;

  printf("opening file %s\n", dataset);
  if ((input = fopen(dataset, "r")) == NULL){printf("input file %s does not exists\n", dataset); exit(1);}        // open the input data
  if (fgets(one_line, sizeof(one_line), input) == NULL){printf("cannot read from the input file %s\n", dataset); exit(1);}
  flag = sscanf(one_line, "#%d#%d#%d#%d#\n", &U, &O, &L, &TMAX);
  if (flag != 4){printf("failed to parse the header line %s (format #number_of_users#number_of_objects#number_of_links#range_of_time_stamps# expected)\n", one_line); exit(1);}
  if ((udeg = calloc(U, sizeof(int))) == NULL){printf("error allocating the user degree array\n"); exit(1);}
  if ((discov = calloc(U, sizeof(int))) == NULL){printf("error allocating the number of discoveries array\n"); exit(1);}
  if ((probNull = calloc(U, sizeof(double))) == NULL){printf("error allocating the probNull array\n"); exit(1);}
  if ((surp = calloc(U, sizeof(double))) == NULL){printf("error allocating the surprisal array\n"); exit(1);}
  if ((sum_k_collect = calloc(U, sizeof(int))) == NULL){printf("error allocating the sum_k_collect array\n"); exit(1);}
  if ((intop = calloc(U, sizeof(int))) == NULL){printf("error allocating the intop array\n"); exit(1);}
  if ((odeg = calloc(O, sizeof(int))) == NULL){printf("error allocating the object degree array\n"); exit(1);}
  if ((odegnow = calloc(O, sizeof(int))) == NULL){printf("error allocating the current object degree array\n"); exit(1);}
  if ((odeginc = calloc(O, sizeof(int))) == NULL){printf("error allocating the degree increment array\n"); exit(1);}
  if ((tau = calloc(O, sizeof(int))) == NULL){printf("error allocating the tau array\n"); exit(1);}
  for (a = 0; a < O; a++) tau[a] = -1;                                          // mark all tau values as unset
  if ((data = calloc(L, sizeof(datastruct))) == NULL){printf("error allocating the data\n"); exit(1);}

  counter = 0;
  while (fgets(one_line, sizeof(one_line), input) != NULL)
    if (one_line[0] != '#')
    {
      flag = sscanf(one_line, "%d\t%d\t%d\n", &user, &object, &time);
      if (flag != 3){printf("failed to parse line %s\n", one_line); exit(1);}
      if (first_line == 0 && time < prev_time){
        printf("line %d\t%d\t%d: time stamp is smaller than in the previous line\n", user, object, time);
        printf("halt: entries in the input file are expected to be sorted from the oldest to the most recent!\n");
        exit(1);}
      data[counter].user = user; data[counter].object = object; data[counter].time = time;
      if (tau[object] == -1) tau[object] = time;
      udeg[user]++; odeg[object]++; counter++;
      prev_time = time; first_line = 0;
    }
  fclose(input);
  printf("%d entries read from the input file (%d users, %d objects)\n", counter, U, O);
  if (counter != L){printf("halt: number of entries differs from the file header specification\n"); exit(1);}

  for (i = 0; i < U; i++)
    if (udeg[i] > maxUdeg) maxUdeg = udeg[i];
  for (a = 0; a < O; a++)
    if (odeg[a] > maxOdeg) maxOdeg = odeg[a];
  mindegD = 0;                                          // find the final degree which at most fracD of all objects have
  counter = O;                                          // objects with degree mindegD and more are considered as target objects for which discoveries are awarded
  while (counter > (int)(fracD * O)){
    for (a = 0; a < O; a++)
      if (odeg[a] == mindegD) counter--;
    mindegD++;
  }
  printf("the average object degree is %.2f, maximal object degree %d\n", (float)L / O, maxOdeg);
  printf("the average user degree is %.2f, maximal user degree %d\n", (float)L / U, maxUdeg);
  printf("\n%.2f%% most popular objects are there to be discovered; first %d users who connect to them are awarded a discovery\n", 100 * fracD, degD);
  printf("there are %d objects (%.2f%%) with final degree %d and more\n", counter, 100 * (float)counter / O, mindegD);

  timenow = 0;                                          // count the number of discoveries by each user
  numD = 0;
  for (n = 0; n < L; n++){
    if (data[n].time > timenow){
      for (a = 0; a < O; a++){odegnow[a] += odeginc[a]; odeginc[a] = 0;}
      timenow = data[n].time;
    }
    odeginc[data[n].object]++;
    sum_k_collect[data[n].user] += odegnow[data[n].object];
    if (odegnow[data[n].object] < degD && odeg[data[n].object] >= mindegD){
      discov[data[n].user]++;
      numD++;
    }
  }
  free(odegnow); free(odeginc);
  pD = (float)numD / L;
  printf("there are %d discoveries in total -> under the null hypothesis, each link has probability %.2f%% to be a discovery\n", numD, 100 * pD);

  sprintf(oname, "data_analysis/surprisal-user_details-%s", dataset);                   // save information about all users with at least one discovery (user ID, degree, number of discoveries, and surprisal)
  output = fopen(oname, "w");
  fprintf(output, "# degD = %d, fD = %.2e, pD = %.4e\n", degD, fracD, pD);
  fprintf(output, "# user\tdegree\tdiscov\tsurprisal\n");
  for (i = 0; i < U; i++){                                                              // compute the probability of the null hypothesis for each user
    probNull[i] = get_Null_prob(discov[i], udeg[i], pD);                                // probability of at least discov[i] discoveries in udeg[i] attempts
    if (probNull[i] == 0){                                                              // underflow in the probability computation
      printf("the null hypothesis probability too low (underflow) for user %d with degree %d and %d discoveries - surprisal is set to 1e6\n", i, udeg[i], discov[i]);
      surp[i] = 1e6;
    }
    else surp[i] = -log(probNull[i]);
    if (discov[i] > 0) fprintf(output, "%d\t%d\t%d\t%.4f\n", i, udeg[i], discov[i], surp[i]);
  }
  fclose(output);
  
  printf("\ntop %d users who have made the most discoveries\n", TOPD);
  printf("rank\tuser\tdiscov\tdegree\tDexpect\tDratio\tprobNull\tsurprisal\tE(k_collect)\n");
  n = 0;
  while(n < TOPD){
    bestD = 0;
    while (intop[bestD] == 1) bestD++;                                                  // find the first user who has not been chosen yet and first set him as the best one
    for (i = 0; i < U; i++)                                                             // go over all users and find whether there is not a better one
      if (discov[i] > bestD && intop[i] == 0) bestD = discov[i];
    counterD = 0;
    for (i = 0; i < U; i++)
      if (discov[i] == bestD){
        counterD++;
        if (bestD > 0)
          printf("%d/%d\t%d\t%d\t%d\t%.2f\t%.2f\t%.2e\t%.2f\t\t%.2f\n", n + 1, counterD, i, discov[i], udeg[i], pD * udeg[i], discov[i] / (pD * udeg[i]), probNull[i], surp[i], (float)sum_k_collect[i] / udeg[i]);
        intop[i] = 1;
      }
    n += counterD;
  }

  printf("\ntop %d users with the highest surprisal\n", TOPD);
  printf("rank\tuser\tdiscov\tdegree\tDexpect\tDratio\tprobNull\tsurprisal\tE(k_collect)\n");
  n = 0;
  for (i = 0; i < U; i++) intop[i] = 0;
  while (n < TOPD){
    bestD = 0;
    while (intop[bestD] == 1) bestD++;                                                  // find the first user who has not been chosen yet and first set him as the best one
    for (i = 0; i < U; i++)                                                             // go over all users and find whether there is not a better one
      if (surp[i] > surp[bestD] && intop[i] == 0) bestD = i;
    counterD = 0;
    for (i = 0; i < U; i++)
      if (fabs(surp[i] - surp[bestD]) < EPSILON)
      {
        counterD++;
        if (surp[bestD] > 0)
          printf("%d/%d\t%d\t%d\t%d\t%.2f\t%.2f\t%.2e\t%.2f\t\t%.2f\n", n + 1, counterD, i, discov[i], udeg[i], pD * udeg[i], discov[i] / (pD * udeg[i]), probNull[i], surp[i], (float)sum_k_collect[i] / udeg[i]);
        intop[i] = 1;
      }
    n += counterD;
  }
  printf("\n");
  free(sum_k_collect); free(intop);
  
  if ((used = calloc(U, sizeof(int))) == NULL){printf("error allocating the used array\n"); exit(1);}
  sprintf(oname, "data_analysis/surprisal-real_profile-%s.dat", dataset);
  output = fopen(oname, "w");
  fprintf(output, "# degD = %d, fD = %.2e, pD = %.4e\n", degD, fracD, pD);
  fprintf(output, "# rank\tsurprisal\tdegree\n");
  prev_val = 0;
  for (n = 0; n < U; n++){                                      // save all real surprisal values
    max_s = -1; whomax = -1;
    for (j = 0; j < U; j++)
      if (surp[j] > max_s && used[j] == 0){max_s = surp[j]; whomax = j;}
    if (max_s != prev_val || compact_real_profile == 0){
      fprintf(output, "%d\t%.4f\n", n + 1, max_s);
      prev_val = max_s;
    }
    if (max_s == 0) break;                                      // stop once we found zero surprisal value
    used[whomax] = 1;
  }
  fclose(output);
  free(used);
}


int main(int argc, char *argv[])
{
  double avg_bootstrap_surprisal;                       // average maximal surprisal in bootstrap realizations
  int num_discoverers;                                  // number of users in the data whose surprisal values are higher than avg_bootstrap_surprisal
  int i, flag;

  init();                                               // RNG and directory initialization

  if (argc == 1){print_usage(); exit(1);}
  if (argc >= 3){
    flag = sscanf(argv[2], "%d", &degD);
    if (flag != 1){printf("failed to parse the input parameter %s\n", argv[2]); print_usage(); exit(1);}
    if (argc >= 4){
      flag = sscanf(argv[3], "%lf", &fracD);
      if (flag != 1){printf("failed to parse the input parameter %s\n", argv[3]); print_usage(); exit(1);}
    }
  }

  load_data(argv[1]);
  avg_bootstrap_surprisal = do_bootstrap(argv[1]);      // get the average highest surprisal in bootstrapped data
  printf("mean best surprisal in bootstrapped data is %.2f (averaged over %d realizations)\n", avg_bootstrap_surprisal, REP_BOOT);
  num_discoverers = 0;
  for (i = 0; i < U; i++)
    if (surp[i] > avg_bootstrap_surprisal) num_discoverers++;
  printf("in real data, %d users have surprisal higher than %.2f\n", num_discoverers, avg_bootstrap_surprisal);
  
  free(udeg); free(discov); free(probNull); free(surp); free(odeg); free(tau); free(data);
  gsl_rng_free(r);
  return(0);
}
