/************************************************************\
* numerical simulation of a bipartite growing network where  *
* users are of distinct ability and objects are of distinct  *
* quality; time decay is employed together with preferential *
* attachment; see the corresponding paper at                 *
* http://arxiv.org/abs/1509.01477                            *
* the GSL library is used for random numbers and special     *
* functions                                                  *
* matus.medo@unifr.ch, June 2016                             *
\************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/stat.h>                   // for directories checking
#include <gsl/gsl_rng.h>                // GSL random number generator
#include <gsl/gsl_sf.h>                 // GSL library with special functions
#include <gsl/gsl_sf_exp.h>             // GSL exponential function
#include <gsl/gsl_sf_log.h>             // GSL logarithm
#include <gsl/gsl_randist.h>            // GSL library with probability density functions
#include <gsl/gsl_statistics.h>         // GSL library with statistics


#define REP             10              // how many network realizations to create for each parameter setting
#define Oi              50              // initial number of objects
#define lifetime        1e3             // object lifetime (aging has exponential form exp(-"object age" / lifetime))
#define OQexp           3.0             // exponent of the power law distribution of object quality
#define eta             0.01            // target network density
#define add             1               // additive term in preferential attachment


int U, UabP, O, Of, L;                  // number of users, user ability parameter, current number of objects, final number of objects, and the final number of links
double newO_period;                     // how often a new object is introduced in the system
double *ab, *act;                       // user ability and user activity
int *udeg;                              // user degree values
int *tau, *odeg;                        // object birth time and object degree values
double *Q;                              // object quality (fitness) values
double *oweight;                        // object weights in the attachment process
char **A;                               // network adjacency matrix U * Of
int **e;                                // link list L * 2 (elements are: from-to)
const gsl_rng_type *gslT;               // for the GSL random number generator
gsl_rng *gslr;                          // for the GSL random number generator


void init(int seed){
  int i;
  struct stat st;

  printf("============================================================================================\n");
  printf("starting network simulation with %d users and %d objects (%d users with high ability)\n", U, Of, UabP);
  printf("============================================================================================\n\n");

  gsl_rng_env_setup();
  gsl_rng_default_seed = seed;
  gslT=gsl_rng_default;
  gslr=gsl_rng_alloc(gslT);
  gsl_set_error_handler_off();                  // to prevent underflows from gsl_sf_exp from halting the program

  L = (int)(eta * U * Of);                      // number of links
  newO_period = (double)L / (Of - Oi + 1);      // how often a new object is introduced in the system

  if ((A = malloc(U * sizeof(char*))) == NULL){printf("init: error allocating the adjacency matrix pointers\n"); exit(1);}
  for (i = 0; i < U; i ++)
    if ((A[i] = malloc(Of * sizeof(char))) == NULL){printf("init: error allocating the adjacency matrix vectors\n"); exit(1);}
  if ((e = malloc(L * sizeof(int*))) == NULL){printf("init: error allocating the link list pointers\n"); exit(1);}
  for (i = 0; i < L; i++)
    if ((e[i] = malloc(2 * sizeof(int))) == NULL){printf("init: error allocating the link list vectors\n"); exit(1);}
  if ((ab = malloc(U * sizeof(double))) == NULL){printf("init: error allocating the user ability array\n"); exit(1);}
  if ((act = malloc(U * sizeof(double))) == NULL){printf("init: error allocating the user activity array\n"); exit(1);}
  if ((udeg = malloc(U * sizeof(int))) == NULL){printf("init: error allocating the user degree array\n"); exit(1);}
  if ((Q = malloc(Of * sizeof(double))) == NULL){printf("init: error allocating the object quality array\n"); exit(1);}
  if ((odeg = malloc(Of * sizeof(int))) == NULL){printf("init: error allocating the object degree array\n"); exit(1);}
  if ((tau = malloc(Of * sizeof(int))) == NULL){printf("init: error allocating the object birth time array\n"); exit(1);}
  if ((oweight = malloc(Of * sizeof(double))) == NULL){printf("init: error allocating the object weight array\n"); exit(1);}

  if (stat("data", &st) != 0) mkdir("data", 0777);      // if it does not exist already, create the output directory for the generated networks
}


void init_realization(){
  int i, a, n, candidate;
  double tot_act;

  for (i = 0; i < U; i++)               // empty the adjacency matrix
    for (a = 0; a < Of; a++)
      A[i][a] = 0;
  tot_act = 0;                          // total activity counter (used to eventually normalize the activity vector)
  for (i = 0; i < U; i++){
    ab[i] = 0;
    act[i] = gsl_rng_uniform(gslr);     // user activity values are uniformly distributed in [0, 1) (this can be changed)
    tot_act += act[i];
    udeg[i] = 0;
  }
  for (i = 0; i < U; i++)               // normalizes user activity (useful for the random choice of user by activity)
    act[i] /= tot_act;
  n = 0;
  while (n < UabP){                     // set ability one for UabP users
    candidate = (int)(U * gsl_rng_uniform(gslr));
    if (ab[candidate] == 0){ab[candidate] = 1; n++;}
  }

  for (a = 0; a < Of; a++){
    Q[a] = pow(1 / gsl_rng_uniform(gslr), 1 / (OQexp - 1));     // set object quality/fitness values
    tau[a] = -1;                                                // for double-checking, the birth time of all objects is initially set to -1
    odeg[a] = 0;                                                // objects have initally no links
  }
  for (a = 0; a < Oi; a++) tau[a] = 0;  // the initially present objects appear have birth time 0
}


int choose(int LENGTH, double *weight){ // choose an integer from the range [0, LENGTH) with weights stored in the weight vector
  double val = gsl_rng_uniform(gslr);
  int n = LENGTH;

  while (val > 0){
    n--;
    if (n == 0) return(0);
    val -= weight[n];
  }
  return n;
}


void save_it(int realization){          // saves the generated network and some basic statistics
  int i, a, n;
  char oname[100];
  FILE *output;

  sprintf(oname, "data/artif_net-%d-%d-%d-%d-user_properties.dat", U, UabP, Of, realization);   // save user properties
  output = fopen(oname, "w");
  fprintf(output, "#user\tact\t\tdegree\tability\n");
  for (i = 0; i < U; i++)
    fprintf(output, "%d\t%.10f\t%d\t%.4f\n", i, act[i], udeg[i], ab[i]);
  fclose(output);

  sprintf(oname, "data/artif_net-%d-%d-%d-%d-object_properties.dat", U, UabP, Of, realization); // save object properties
  output = fopen(oname, "w");
  fprintf(output, "#object\ttau\tdegree\tfitness\n");
  for (a = 0; a < Of; a++)
    fprintf(output, "%d\t%d\t%d\t%.4f\n", a, tau[a], odeg[a], Q[a]);
  fclose(output);

  sprintf(oname, "data/artif_net-%d-%d-%d-%d.dat", U, UabP, Of, realization);                   // save the network's links
  output = fopen(oname, "w");
  fprintf(output, "#%d#%d#%d#%d#\n", U, Of, L, L);
  fprintf(output, "# O_i = %d, sparsity = %.4f, A = %.1f, lifetime = %.0f\n", Oi, eta, (double)add, lifetime);
  fprintf(output, "#user\tobject\ttime\n");
  for (n = 0; n < L; n++)
    fprintf(output, "%d\t%d\t%d\n", e[n][0], e[n][1], n);
  fclose(output);
}


int main(int argc, char *argv[]){
  char flag;
  int o, i, a, links, added_objects, reporting_period;
  double weight_sum;

  if (argc != 4){                                               // parse the input parameters and create the corresponding arrays
    printf("three parameters needed: number of users, number of high-ability users, and the final number of objects\n"); exit(1);}
  flag = sscanf(argv[1], "%d", &U);
  if (flag != 1){printf("failed to parse parameter %s\n", argv[1]); exit(1);}
  flag = sscanf(argv[2], "%d", &UabP);
  if (flag != 1){printf("failed to parse parameter %s\n", argv[2]); exit(1);}
  flag = sscanf(argv[3], "%d", &Of);
  if (flag != 1){printf("failed to parse parameter %s\n", argv[3]); exit(1);}
  init(0);

  for (o = 0; o < REP; o++){                                    // generate the required number of independent network realizations
    printf("realization %d - progress:", o); fflush(stdout);
    init_realization();
    O = Oi; links = 0; added_objects = 0; reporting_period = L / 100;
    while (links < L){
      if (links > newO_period * (added_objects + 1) && O < Of){ // if it's time, add a new object
        tau[O] = links; added_objects++; O++;}
      i = choose(U, act);
      for (a = 0; a < Of; a++) oweight[a] = 0;
      weight_sum = 0;
      if (ab[i] == 0){                                          // zero ability users choose by popularity
        for (a = 0; a < O; a++){
          oweight[a] = (odeg[a] + 1) * gsl_sf_exp(-(double)(links - tau[a]) / lifetime);
          weight_sum += oweight[a];
        }
        for (a = 0; a < O; a++) oweight[a] /= weight_sum;
      }
      else if (ab[i] == 1){                                     // users with ability one choose by quality
        for (a = 0; a < O; a++){
          oweight[a] = Q[a] * gsl_sf_exp(-(double)(links - tau[a]) / lifetime);
          weight_sum += oweight[a];
        }
        for (a = 0; a < O; a++) oweight[a] /= weight_sum;
      }
      else{printf("oops: user %d has ability %.4f which matches neither 0, neither 1\n", i, ab[i]); exit(1);}
      do{
        a = choose(O, oweight);                                 // based on the computed weight vector, choose an object
      } while (A[i][a] == 1);                                   // a user cannot connect with an object that they have already connected with
      A[i][a] = 1;                                              // make the corresponding changes of all arrays
      e[links][0] = i;
      e[links][1] = a;
      links ++;
      udeg[i] ++;
      odeg[a] ++;
      if (links % reporting_period == 0){
        printf(" %d%%", links / reporting_period); fflush(stdout);}
    }
    save_it(o);
    printf("\nnetwork %d generated and saved\n", o);
  }

  for (i = 0; i < U; i++) free(A[i]);                           // free all allocated arrays
  free(A);
  for (i = 0; i < L; i++) free(e[i]);
  free(e);
  free(ab); free(act); free(udeg); free(Q); free(odeg); free(tau); free(oweight);
  gsl_rng_free(gslr);

  printf("\nprogram finished\n");
  return(0);
}
